
#ifdef NCURSES
  #include <curses.h>
#else
  #include <curses.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/wait.h>

#define BLANCO 0

/* Globales */

int num_nodos=0,	/*  Variable que contabilizara el numero    */
			/*  total de nodos generados.               */

    iguales=0;          /*  Variable que contabilizara el numero    */
			/*  de nodos repetidos.                     */

/* **************************************************************** */
/*  Obtiene un entero entre "min" y "max", tomando como cero si se  */
/*  le da uno fuera del rango, o se pulsa una tecla no numerica.    */
/*  Si dentro del rango existen numeros de dos cifras, espera por   */
/*  ENTER para indicar fin del numero.                              */
/* **************************************************************** */

int getint(int min,int max,int numComponentes)
   {
     int numero=0,       /*     Almacenara el numero obtenido.      */

     n,       /* Indicara el numero de pulsaciones que se espera.   */

     i=0,     /* Indicara si se esta esperando por unidades (0),    */
	      /* decenas (1), o centenas (3).                       */
     tecla;

     printw("  \b\b");

     n=(numComponentes<10)?2:3; /* Si solo se esperan numeros de 1 cifra, se */
		       /* espera 1 pulsacion, sino 3 (2 y ENTER).   */

     do
       {
	 /* Si se trata de la ultima pulsacion, no la imprime en    */
	 /* pantalla, sino si.                                      */

/*
	 if (i<n-1)	echo();
	 else	 
*/	 
	 noecho();
	 tecla=getch();
	 if (tecla!=10)   
	 {
	   numero=(tecla-0x30)+10*i*numero;
	   printw("%c",tecla);
           refresh();
	 }
         echo();
	 i++;
       }
     while (i<n && tecla!=10);

     if (numero<min || numero>max) numero=0;

     return(numero);
   }

/* **************************************************************** */
/*  Dibuja en pantalla las lineas del puzzle.                       */
/* **************************************************************** */


void dib_puzle(int dh, int dv, int numFilas, int numCols)
    {
      int i,j;

      j=1+dv;

      for (i=0;i<numCols+1;i++)
         {
           move(j,2+dh);      
           hline('-',numFilas*4+1);
           j+=2;
         }

      j=2+dh;
      for (i=0;i<numFilas+1;i++)
         {
           move(1+dv,j);
           vline('|',numCols*2+1);
           j+=4;
         }

      refresh();
    }

/* **************************************************************** */
/*  Pide un estado, posicion a posicion, indicando en pantalla cual */
/*  se solicita en cada momento. El estado se guardara en "vector". */
/* **************************************************************** */


void pedir_vector(char * texto ,int dh ,int dv,int vector[],int numFilas,
                   int numCols)
    {
      int i,x,y,fila,col,numero;

      move(2,dh-2);       
      hline('-',20);
      mvprintw(1,dh-2,"%s (%d,%d)",texto,1,1);
      dib_puzle(dh,dv,numFilas,numCols);

      for (i=0;i<numFilas*numCols;i++)
         {
           x=(i / numFilas)+1;
           y=(i % numFilas)+1;
           mvprintw(1,dh+12," (%d,%d)",x,y);
           fila=(y*4)+dh;
           col=(x*2)+dv;
           move(col,fila);
           numero=getint(0,(numFilas*numCols)-1, numFilas*numCols);
           move(col,fila);
           vector[i]=numero;
           printw("  \b\b%d",numero);
           refresh();
         }
    }

/* **************************************************************** */
/*  Mira si el estado almacenado en "vector" contiene todos los va- */
/*  lores entre 0 y TOTAL-1.                                        */
/* **************************************************************** */


int comprobar_vector(int vector[], int numeroComponentes)
   {
     int i,j;

     for (i=0;i<numeroComponentes;i++)
     {
        j=0;
        while  (j<numeroComponentes && vector[j]!=i) 
          j++;
        if (j==numeroComponentes) 
          return(-1);
     }
     return(0);
   }

/****************************************************************************/
/*  Partiendo del numero de filas y columnas centra el puzzle en la pantalla*/
/****************************************************************************/
void conseguirCoordenadas(int *vertical, int* horizontal, int numero, int filas,
                          int columnas)
{
     
      *vertical = (24 - (3*filas)) / 2;
      *horizontal = (40 - 5*columnas) / 2 + numero*40;
      
}

/* **************************************************************** */
/*  Pide los estados inicial y final, y la heuristica.              */
/* **************************************************************** */


void entrada(int **inicio,int **final, int *algoritmo, int * heuristica, int * filas, int* cols)
    {

      WINDOW * stdscr;
      int unDigito=9,  vertical, horizontal;

      stdscr=initscr();

      do 
      {
        mvprintw(12,10,"Introduce numero de filas (Max 6): " );
        *filas = getint(2,6,unDigito);
        mvprintw(12,45,"%d",*filas);
      }
      while (*filas<2 || *filas>6);
    
      do
      {
        mvprintw(13,10,"Introduce numero de columnas (Max 6): ");
        *cols = getint(2,6,unDigito);
        mvprintw(20,48,"%d",*cols);
      }
      while (*cols<2 || *cols>6);
               
      erase();
      
      *inicio = (int *)malloc(sizeof(int)*(*filas)*(*cols));
      *final = (int *)malloc(sizeof(int)*(*filas)*(*cols));
      conseguirCoordenadas(&vertical, &horizontal,0, *filas, *cols); 
      do
      {
         pedir_vector("Estado inicial",horizontal,vertical,*inicio, *filas, 
          *cols);
      }
      while (comprobar_vector(*inicio, (*filas)*(*cols))!=0);
      conseguirCoordenadas(&vertical, &horizontal,1, *filas, *cols); 
      do
      {
         pedir_vector("Estado final  ",horizontal, vertical,*final, *filas, 
          *cols);
      }
      while (comprobar_vector(*final, (*filas)*(*cols))!=0);

      erase();
      do
      {
          mvprintw(10,20,"0 - DFID");
          mvprintw(11,20,"1 - A*");
          mvprintw(12,20,"2 - IDA*");
          mvprintw(13,20,"3 - SMA*");
          mvprintw(16,10,"Introduzca el algoritmo (0-3) ");
          *algoritmo = getint(0,3,unDigito);
          mvprintw(16,40,"%d",*algoritmo);
          refresh();
      }
      while (*algoritmo<0 || *algoritmo>3);
      
      erase();
      do
      {
       
          mvprintw(10,20,"0 - Distancias");
          mvprintw(11,20,"1 - Aciertos");
          mvprintw(12,20,"2 - Anchura");
          mvprintw(15,10,"Introduzca la heuristica(0-2) ");
          *heuristica = getint(0,2,unDigito);
          mvprintw(15,40,"%d",*heuristica);
          refresh();
      }
      while (*heuristica<0 || *heuristica>2);
      erase ();
      
    }



/* **************************************************************** */
/*  Imprime el estado "vector" en pantalla en forma de puzzle.      */
/* **************************************************************** */


void imp_vector(int dh ,int dv,int vector[], int numFilas, int numCols)
    {
      int i,x,y,fila,col;

      dib_puzle(dh,dv,numFilas,numCols);
                   		/*  Dibuja las lineas del puzzle.   */
      /* Coloca en cada casilla el numero correspondiente.          */
      for (i=0;i<numFilas*numCols;i++)
         {
           x=(i / numFilas)+1;
           y=(i % numFilas)+1;
           fila=(y*4)+dh;
           col=(x*2)+dv;
           if (vector[i]!=0)      
             mvprintw(col,fila,"%d",vector[i]);
           else 
             mvprintw(col,fila,"  ");
           refresh();
         }
      refresh();
     }

/* **************************************************************** */
/* **************************************************************** */
/* **************************************************************** */

int buscarBlanco(int inicial[], int numFilas, int numCols)
{
   int i, numElem;
   
     numElem = numFilas * numCols; 
     for (i=0; i < numElem; i++)
     {
       if (inicial[i] == BLANCO)
         return i;
     }  
     exit (1);
}


void aplicarMovimiento(int inicial[], char movimiento, int numFilas, 
                       int numCols)
{
   int posBlanco;
   
     posBlanco = buscarBlanco(inicial, numFilas, numCols);
     switch (movimiento) 
     {
        case 'E' : inicial[posBlanco] = inicial[posBlanco+1];
                   inicial[posBlanco+1] = BLANCO;
                   break;
                   
        case 'O' : inicial[posBlanco] = inicial[posBlanco-1];
                   inicial[posBlanco-1] = BLANCO;
                   break;
                   
        case 'N' : inicial[posBlanco] = inicial[posBlanco-numCols];
                   inicial[posBlanco-numCols] = BLANCO;
                   break;
                   
        case 'S' : inicial[posBlanco] = inicial[posBlanco+numCols];
                   inicial[posBlanco+numCols] = BLANCO;
                   break;
 /*       default : printf("Y una mierda\n");
                  refresh();  
 */
     }
}

/* **************************************************************** */
/*  Halla a partir del puntero al nodo meta, mediante los punteros   */
/*  padre, el camino desde el inicial al final, almacenandolo en    */
/*  una lista, que luego recorrera, imprimiendo uno a uno.          */
/* **************************************************************** */


int Salida (int *estado, int numFilas, int numCols)
{
   	
   	        /* Variable que acumulara la longitud del camino. */
		  /* Como longitud se toma el numero de movimientos.*/
   FILE *df;
   char movimiento;

      erase();      
      df = fopen ("solucion", "r");
      if (df!=NULL)
      {
        
        imp_vector (30,5,estado,numFilas,numCols);
        move(1,1);
        fread(&movimiento, sizeof(char), 1, df);
        aplicarMovimiento(estado,movimiento,numFilas,numCols);

        while (!feof(df))
        {
 
           while (getch()!=10);
          
           erase ();
	   imp_vector (30,5,estado,numFilas,numCols);
           move(1,1);
	   fread(&movimiento, sizeof(char), 1, df);
	   aplicarMovimiento(estado,movimiento,numFilas,numCols);
        }
        fclose(df);
        return (1);
      /* Imprime los resultados finales. */
      }
      erase();
      return (0);
}



void escribirVector(int vector[], char *nombre, int nFilas, int nCols)
{
   int i, j;
   FILE *df;
      
     df = fopen(nombre, "w");
     fprintf(df, "%dx%d\n", nFilas, nCols);
     for (i=0; i<nFilas; i++)
     {
       for (j=0; j<nCols; j++)
         fprintf(df, "%d\t", vector[i*nCols+j]);
       fprintf(df, "\n");
     }
     fclose(df);
}

void traducirArgumento(int valor, char *argumento)
{
   
     argumento[0] = valor + 0x30;
     argumento[1] = '\0';
}


void leerDatosGenerales( int *nodosTotales, int *generados, int *repes, int *profundidad,
                         int *numMovs, float *tiempo)
{
  FILE *fdatosGenerales;

    fdatosGenerales = fopen("generales","r");
    fscanf(fdatosGenerales,"%d\n%d\n%d\n%d\n%d\n",nodosTotales, generados,
                            repes, profundidad, numMovs);
                                        
    fscanf(fdatosGenerales,"%f ",tiempo);                                    
    fclose(fdatosGenerales);                                   
                                        
}

extern char **environ;

void main ()
{
  int error, solucion;
  int *estadoInicial, *estadoFinal, algoritmo, heuristica, filas, columnas;
  char *argAlgoritmo, *argHeuristica, *cadena;
  int nodosTotales, profundidad, repes, generados, numMovs; 
  float tiempo;
  int childStatus;
   

     entrada(&estadoInicial, &estadoFinal, &algoritmo, &heuristica, &filas, &columnas);
     erase();
     escribirVector(estadoInicial, "inicial", filas, columnas);
     escribirVector(estadoFinal, "final", filas, columnas);
     argAlgoritmo = (char *) malloc(sizeof(char)*2);
     argHeuristica = (char *) malloc(sizeof(char)*2);
     traducirArgumento(algoritmo, argAlgoritmo);
     traducirArgumento(heuristica, argHeuristica);
     cadena = (char *) malloc(sizeof(char) * 100);
     strcpy(cadena, "puzzle");
   
     if (fork()==0)
     {
       error = execle( cadena, "", argAlgoritmo, argHeuristica, NULL, environ);
       exit(0); 
     }
     
       /* Sincronizacion con puzzle.c */
       wait(&childStatus)   ;

       solucion = Salida(estadoInicial, filas, columnas);
       refresh();
       if (solucion)
       {
        leerDatosGenerales(&nodosTotales, &generados, &repes, &profundidad,  
                        &numMovs, &tiempo);
       remove("solucion");
       remove("generales");
              
       remove("inicial");
       remove("final");
       
       mvprintw(17,10,"Nodos generados           = %d",nodosTotales);
       mvprintw(18,10,"Nodos nuevos              = %d",generados);
       mvprintw(19,10,"Nodos repetidos           = %d",repes);
       mvprintw(20,10,"Profundidad               = %d",profundidad);
       mvprintw(21,10,"Longitud del camino       = %d",numMovs);
       mvprintw(22,10,"Tiempo                    = %.3f",tiempo);
       
       }
     
     else
       mvprintw(18,10,"Los estados pertenecen a configuraciones distintas ");
     refresh();
     getch();                   
     endwin();
}




