#!/usr/local/bin/wish -f

###############################
#  Procedimientos auxiliares  #
###############################

proc moverPuzzle {w num} {
    global fila columna filas columnas espacio

    set tmp $fila($espacio)
    set fila($espacio) $fila($num)
    set fila($num) $tmp
    set tmp $columna($espacio)
    set columna($espacio) $columna($num)
    set columna($num) $tmp
    set posy [expr $fila($num)/double($filas)]
    set posx [expr $columna($num)/double($columnas)]
    place $w.$num -relx $posx -rely $posy
}


proc moverPuzzleRestr {w puzzle direccion} {
    upvar $puzzle puz
    global fila columna filas columnas espacio pasos

    switch $direccion {
      N {
         set f [expr $fila($espacio)-1]
         set c $columna($espacio)
         set fpaso [expr 1.0/($filas*$pasos)]
         set cpaso 0
      } 
      S {
         set f [expr $fila($espacio)+1]
         set c $columna($espacio)
         set fpaso [expr -1.0/($filas*$pasos)]
         set cpaso 0
      }
      E {
         set f $fila($espacio)
         set c [expr $columna($espacio)+1]
         set fpaso 0
         set cpaso [expr -1.0/($columnas*$pasos)]
      }       
      O {
         set f $fila($espacio)
         set c [expr $columna($espacio)-1]
         set fpaso 0
         set cpaso [expr 1.0/($columnas*$pasos)]
      } 
    }
    
    # Intercambiamos la posicion con el espacio
    set num $puz($f,$c)
    set puz($f,$c) $espacio
    set puz($fila($espacio),$columna($espacio)) $num

    # Intercambiamos la posicion con el espacio (otra)
    set tmp $fila($espacio)
    set fila($espacio) $fila($num)
    set fila($num) $tmp
    set tmp $columna($espacio)
    set columna($espacio) $columna($num)
    set columna($num) $tmp

    set fcont [expr $f/double($filas)]
    set ccont [expr $c/double($columnas)]

    $w.$num flash
    for {set paso 0} {$paso <= $pasos} {set paso [expr $paso+1]} {
      place $w.$num -relx $ccont -rely $fcont
      update idletasks
      set fcont [expr $fcont + $fpaso]
      set ccont [expr $ccont + $cpaso]
    }
}



proc verPuzzle {w} {
  global fila columna filas columnas espacio
  set num 0
  for {set f 0} {$f < $filas} {set f [expr $f+1]} {
    for {set c 0} {$c < $columnas} {set c [expr $c+1]} {
      if {$num != $espacio} { 
        set posy [expr $fila($num)/double($filas)]
        set posx [expr $columna($num)/double($columnas)]
        place $w.$num -relx $posx -rely $posy \
              -relwidth [expr 1.0/$columnas]  \
              -relheight [expr 1.0/$filas]
      }
      set num [expr $num+1] 
    }
  }    
}



proc crearPuzzle {w} {
  global fila columna filas columnas espacio
  set num 0
  set tFont 240
  for {set f 0} {$f < $filas} {set f [expr $f+1]} {
    for {set c 0} {$c < $columnas} {set c [expr $c+1]} {
      set fila($num) $f
      set columna($num) $c
      if {$num != $espacio} { 
        button $w.$num       \
              -relief raised \
              -text $num     \
              -highlightthickness 0 \
              -command "moverPuzzle $w $num"  \
              -font "-*-Helvetica-Medium-R-Normal--*-$tFont-*-*-*-*-*-*"
        set posy [expr $fila($num)/double($filas)]
        set posx [expr $columna($num)/double($columnas)]
        place $w.$num     \
              -relx $posx \
              -rely $posy \
              -relwidth [expr 1.0/$columnas]  \
              -relheight [expr 1.0/$filas]
      }
      set num [expr $num+1] 
    }
  }    
}


proc destruirPuzzle {w} {
  global espacio filas columnas
  set num 0
  for {set f 0} {$f < $filas} {set f [expr $f+1]} {
    for {set c 0} {$c < $columnas} {set c [expr $c+1]} {
      if {$num != $espacio} { 
        destroy $w.$num 
      }
      set num [expr $num+1]
    }
  }    
}


proc grabarPuzzle {fich} {
  global fila columna filas columnas espacio
  set fd [open $fich w]
  puts $fd [format "%dx%d" $filas $columnas]

  # Calculamos las posiciones
  set num 0
  for {set f 0} {$f < $filas} {set f [expr $f+1]} {
    for {set c 0} {$c < $columnas} {set c [expr $c+1]} {
      set puzzle($fila($num),$columna($num)) $num 
      set num [expr $num+1]
    }
  }
  
  # Escribimos en el fichero
  for {set f 0} {$f < $filas} {set f [expr $f+1]} {
    for {set c 0} {$c < $columnas} {set c [expr $c+1]} {
      puts -nonewline $fd [format "%d\t" $puzzle($f,$c)]
      set num [expr $num+1]
    }
    puts $fd ""
  }
  flush $fd
  close $fd
}



proc leerPuzzle {w fich} {
  global fila columna filas columnas
  set fd [open $fich r]
  scan [gets $fd] "%dx%d" tmpFilas tmpColumnas
  
  # Comprobar si la configuracion grabada tiene
  # diferente numero de columnas o filas
  if {$tmpFilas != $filas || $tmpColumnas != $columnas} {
    destruirPuzzle $w
    set filas $tmpFilas
    set columnas $tmpColumnas
    crearPuzzle $w
  }

  # Leemos del fichero
  for {set f 0} {$f < $filas} {set f [expr $f+1]} {
    gets $fd strFila
    for {set c 0} {$c < $columnas} {set c [expr $c+1]} {
      scan $strFila "%d" tmp
      set puzzle($f,$c) $tmp
      set strFila [string trimleft $strFila 0123456789]
      set strFila [string trimleft $strFila "\t"]
    }
  }

  # Calculamos las posiciones
  for {set f 0} {$f < $filas} {set f [expr $f+1]} {
    for {set c 0} {$c < $columnas} {set c [expr $c+1]} {
      set fila($puzzle($f,$c)) $f
      set columna($puzzle($f,$c)) $c
    }
  }
  verPuzzle $w
  close $fd
}


proc verSolucion {w ops fini fsol} {
  global fila columna filas columnas espacio nMov cancelar pasos 
  leerPuzzle $w $fini
  verPuzzle $w
  set fd [open $fsol r]

  # Calculamos las posiciones
  set num 0
  for {set f 0} {$f < $filas} {set f [expr $f+1]} {
    for {set c 0} {$c < $columnas} {set c [expr $c+1]} {
      set puzzle($fila($num),$columna($num)) $num
      set num [expr $num+1]
    }
  }
 

  frame .frSol
  scale .frSol.scPasos \
          -orient horizontal \
          -from 1 -to 200 \
          -variable pasos

  set cancelar no   
  frame .frSol.frPaso
  label .frSol.frPaso.lPaso -text "Paso:"
  button .frSol.frPaso.bPaso -text 0 -relief sunken
  pack .frSol.frPaso.lPaso .frSol.frPaso.bPaso \
        -side left -fill x -expand yes

  button .frSol.bCancelar -text "Cancelar" \
          -command {set cancelar yes} 
  pack .frSol.frPaso .frSol.scPasos .frSol.bCancelar -side top -fill x 

  # Cambiar los botones
  pack forget $ops   
  pack .frSol -side bottom -fill x
  update  

  .frSol.bCancelar flash

  set nMovs 1  
  set movimiento [gets $fd]  
  while {![eof $fd] && $cancelar == "no"} {
    .frSol.frPaso.bPaso config -text $nMovs
    moverPuzzleRestr $w puzzle $movimiento   
    update 
    set movimiento [gets $fd]
    set nMovs [expr $nMovs+1]
  }
  .frSol.frPaso.bPaso config -background red
  update
  after 1000
  destroy .frSol
  pack $ops -fill x
  close $fd
}


proc verDatosGenerales {fich} {
  global Directorio
  set fd [open $fich r]
  set Parametros {
    "Nodos Generados:"      
    "Nodos Nuevos:"  
    "Nodos Repetidos:"
    "Maxima profundidad:"
    "Pasos de la solucion:"
    "Tiempo(de usuario):"
  }

  catch {destroy .tl}
  toplevel .tl 
  wm title .tl "Resultados $Directorio"

  frame .tl.frDts 

  frame .tl.frDts.frEtiqs -borderwidth 0.5c
  frame .tl.frDts.frVals -borderwidth 0.5c
  foreach Parametro $Parametros {
      label .tl.frDts.frEtiqs.l$Parametro \
          -text $Parametro \
          -font "-*-Helvetica-Medium-R-Normal--*-240-*-*-*-*-*-*"
      pack .tl.frDts.frEtiqs.l$Parametro \
          -side top \
          -anchor e

      label .tl.frDts.frVals.l$Parametro \
          -text [gets $fd]  \
          -font "-*-Helvetica-Medium-R-Normal--*-240-*-*-*-*-*-*"
      pack .tl.frDts.frVals.l$Parametro \
          -side top \
          -anchor w
  }

  button .tl.bAceptar \
      -text "Aceptar" \
      -command {destroy .tl}
  pack .tl.frDts.frEtiqs .tl.frDts.frVals -side left 
  pack .tl.frDts .tl.bAceptar -side top -fill both -expand yes
  close $fd
}


##############################
#       Principio            #
##############################

option add *Button.relief ridge


set pasos 50
set espacio 0
set fichInicial inicial
set fichFinal final
set fichSolucion solucion
set fichDatosGenerales generales
set queEstado $fichInicial
set tHeur 0
set rm /bin/rm
set Heuristicas {
   "Minimizar Distancias"
   "Maximizar Aciertos"
   "h=0"
 }
set tAlg 0
set Algoritmos {
    "DFID"
    "A*"
    "IDA*"
    "SMA*"
 }
                 

wm title . Puzzle

# Comprobamos los argumentos
set filas 3
set columnas 3
set Directorio 1
if {$argc == 3} { 
   set filas [lindex $argv 0]
   set columnas [lindex $argv 1]
   set Directorio [lindex $argv 2]  
}

catch {exec mkdir $Directorio}
cd $Directorio


# Frame donde colocamos el puzzle
frame .frPuzzle \
        -height "$filas i"   \
        -width "$columnas i" \
        -borderwidth 2 -relief sunken \
        -background #5f9ea0


# Creamos los botones para el puzzle dandoles tamano y posicion
crearPuzzle .frPuzzle
grabarPuzzle inicial
grabarPuzzle final


# Frame donde colocamos todas las opciones
frame .frOps



# Botones de eleccion de estado inicial/final
frame .frOps.frIniFin

radiobutton .frOps.frIniFin.rbInicial  \
        -text Inicial \
        -variable queEstado \
        -value $fichInicial \
        -highlightthickness 0 \
        -command { if {$rAuto == 1} {leerPuzzle .frPuzzle inicial}} \
        -selectcolor yellow 

radiobutton .frOps.frIniFin.rbFinal \
        -text Final \
        -variable queEstado \
        -value $fichFinal    \
        -highlightthickness 0 \
        -command { if {$rAuto == 1} {leerPuzzle .frPuzzle final}} \
        -selectcolor yellow 

pack .frOps.frIniFin.rbInicial -side left -fill x -expand yes
pack .frOps.frIniFin.rbFinal -side right -fill x -expand yes



# Seleccion de refresco automatico de configuracion
checkbutton .frOps.ckAuto \
         -text "Refresco automatico" \
         -highlightthickness 0 \
         -variable rAuto \
         -selectcolor yellow 


# Botones para resolver y ver solucion
frame .frOps.frSoluc \
     -bg blue

button .frOps.frSoluc.bResolver \
      -text "Resolver" \
      -command {
	 catch {exec $rm -f $fichSolucion}
	 catch {exec ../puzzle $tAlg $tHeur}
         if {[file exists $fichSolucion]} {
            .frOps.frSoluc.bVerSolucion config -state normal
            verDatosGenerales $fichDatosGenerales
         update
         for {set n 0} {$n < 4} {incr n} {
            bell
            after 140
         }          
	 } else {
         for {set n 0} {$n < 2} {incr n} {
            bell
            after 140
         }          
         tk_dialog .confDisjuntas "Puzzle" "Estados pertenecientes a configuraciones disjuntas" "" 0 "Aceptar"
	 }         
      }

button .frOps.frSoluc.bVerSolucion \
      -text "Ver solucion" \
      -state disabled      \
      -command {verSolucion .frPuzzle .frOps $fichInicial $fichSolucion}

pack .frOps.frSoluc.bResolver .frOps.frSoluc.bVerSolucion -side left -fill x -expand yes

frame .frOps.frAlgHeur -borderwidth 2 -relief ridge 


frame .frOps.frAlgHeur.frAlg -borderwidth 10 -relief flat

label .frOps.frAlgHeur.frAlg.lAlgoritmo \
      -text "Algoritmos\n" \
      -foreground blue \
    

pack .frOps.frAlgHeur.frAlg.lAlgoritmo 
  

set aValor 0
foreach Algoritmo $Algoritmos {
  radiobutton .frOps.frAlgHeur.frAlg.b$Algoritmo \
      -text $Algoritmo \
      -variable tAlg \
      -highlightthickness 0 \
      -value $aValor \
      -selectcolor yellow
   
  pack .frOps.frAlgHeur.frAlg.b$Algoritmo \
      -side top \
      -anchor w
 set aValor [expr $aValor+1]
}


frame .frOps.frAlgHeur.frHeur -borderwidth 2 -relief flat

label .frOps.frAlgHeur.frHeur.lHeuristica \
      -text "Heuristicas\n\n" \
      -foreground blue \
  
      
pack .frOps.frAlgHeur.frHeur.lHeuristica 

set hValor 0
foreach Heuristica $Heuristicas {
  radiobutton .frOps.frAlgHeur.frHeur.b$Heuristica \
      -text $Heuristica \
      -variable tHeur \
      -highlightthickness 0 \
      -value $hValor \
      -selectcolor yellow 
       
  pack .frOps.frAlgHeur.frHeur.b$Heuristica \
      -side top \
      -anchor w
  set hValor [expr $hValor+1]
}

pack .frOps.frAlgHeur.frAlg \
     -side left -fill x -expand yes
     
pack .frOps.frAlgHeur.frHeur \
     -side right -fill x -expand yes
     

# Botones para leer y guardar configuraciones
frame .frOps.frGuarLeer
button .frOps.frGuarLeer.bGuardar \
      -text "Guardar" \
      -command {
          .frOps.frSoluc.bVerSolucion config -state disabled
          grabarPuzzle $queEstado
       }

button .frOps.frGuarLeer.bLeer \
      -text "Leer" \
      -command {
           leerPuzzle .frPuzzle $queEstado
      }

pack .frOps.frGuarLeer.bLeer \
     .frOps.frGuarLeer.bGuardar \
     -side right -fill x -expand yes
     


button .frOps.bSalir \
      -text "Salir" \
      -highlightthickness 0 \
      -command {         
         cd ..
         catch {exec rm -fr $Directorio}
         exit
       }

pack .frOps.frIniFin \
     .frOps.ckAuto \
     .frOps.frGuarLeer \
     .frOps.frSoluc \
     .frOps.frAlgHeur \
     -side top -fill x   


pack .frOps.bSalir \
     -side top -fill x

pack .frPuzzle -side top -fill both -expand yes
pack .frOps -side bottom -fill x
	


